#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

""":term:`APD` detector modeling class."""

from collections.abc import Mapping
from typing import TYPE_CHECKING

from pyxel.detectors import Detector

if TYPE_CHECKING:
    import pandas as pd

    from pyxel.detectors import APDCharacteristics, APDGeometry, Environment


class APD(Detector):
    """:term:`CMOS`-based detector class containing all detector attributes and data."""

    def __init__(
        self,
        geometry: "APDGeometry",
        environment: "Environment",
        characteristics: "APDCharacteristics",
    ):
        self._geometry: APDGeometry = geometry
        self._characteristics: APDCharacteristics = characteristics

        super().__init__(environment=environment)
        super().reset()

    def __eq__(self, other) -> bool:
        return (
            type(self) is type(other)
            and self.geometry == other.geometry
            and self.environment == other.environment
            and self.characteristics == other.characteristics
            and super().__eq__(other)
        )

    @property
    def geometry(self) -> "APDGeometry":
        """TBW."""
        return self._geometry

    @property
    def characteristics(self) -> "APDCharacteristics":
        """TBW."""
        return self._characteristics

    # TODO: Refactor this
    def to_dict(self) -> Mapping:
        """Convert an instance of `APD` to a `dict`."""
        dct = {
            "version": 1,
            "type": "APD",
            "properties": {
                "geometry": self.geometry.to_dict(),
                "environment": self.environment.to_dict(),
                "characteristics": self.characteristics.to_dict(),
            },
            "data": {
                "photon": None if self._photon is None else self._photon.array.copy(),
                "pixel": None if self._pixel is None else self._pixel.array.copy(),
                "signal": None if self._signal is None else self._signal.array.copy(),
                "image": None if self._image is None else self._image.array.copy(),
                "data": None if self._data is None else self._data.to_dict(),
                "charge": (
                    None
                    if self._charge is None
                    else {
                        "array": self._charge.array.copy(),
                        "frame": self._charge.frame.copy(),
                    }
                ),
                "scene": (
                    None
                    if self._scene is None
                    else {
                        key.replace("/", "#"): value
                        for key, value in self._scene.to_dict().items()
                    }
                ),
            },
        }

        return dct

    # TODO: Refactor this
    @classmethod
    def from_dict(cls, dct: Mapping) -> "APD":
        """Create a new instance of `APD` from a `dict`."""
        # TODO: This is a simplistic implementation. Improve this.
        import numpy as np
        import xarray as xr
        from datatree import DataTree

        from pyxel.data_structure import Scene
        from pyxel.detectors import APDCharacteristics, APDGeometry, Environment

        if dct["type"] != "APD":
            raise ValueError

        if dct["version"] != 1:
            raise ValueError

        properties = dct["properties"]
        geometry = APDGeometry.from_dict(properties["geometry"])
        environment = Environment.from_dict(properties["environment"])
        characteristics = APDCharacteristics.from_dict(properties["characteristics"])

        detector = cls(
            geometry=geometry,
            environment=environment,
            characteristics=characteristics,
        )

        data = dct["data"]

        if "photon" in data:
            detector.photon.array = np.asarray(data["photon"])
        if "pixel" in data:
            detector.pixel.array = np.asarray(data["pixel"])
        if "signal" in data:
            detector.signal.array = np.asarray(data["signal"])
        if "image" in data:
            detector.image.array = np.asarray(data["image"])
        if "data" in data:
            detector._data = DataTree.from_dict(
                {
                    key: xr.Dataset.from_dict(value)
                    for key, value in data["data"].items()
                }
            )
        if "scene" in data and (scene_dct := data["scene"]) is not None:
            detector.scene = Scene.from_dict(
                {key.replace("#", "/"): value for key, value in scene_dct.items()}
            )
        if "charge" in data and data["charge"] is not None:
            charge_dct = data["charge"]
            detector.charge._array = np.asarray(charge_dct["array"])

            new_frame: pd.DataFrame = charge_dct["frame"]
            previous_frame: pd.DataFrame = detector.charge._frame
            detector.charge._frame = new_frame[previous_frame.columns]

        return detector



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

"""Convert scene to photon with aperture model."""

import astropy.units as u
import numpy as np
import xarray as xr
from astropy import wcs
from astropy.coordinates import SkyCoord

from pyxel.data_structure import Scene
from pyxel.detectors import Detector


def extract_wavelength(
    scene: Scene,
    wavelength_band: tuple[float, float],
) -> xr.Dataset:
    """Extract xarray Dataset of Scene for selected wavelength band.

    Parameters
    ----------
    scene : Scene
        Pyxel scene object.
    wavelength_band : float
        Selected wavelength band. Unit: nm.

    Returns
    -------
    selected_data : xr.Dataset

    Examples
    --------
    >>> scene = Scene(...)
    >>> extract_wavelength(scene=scene, wavelength_band=[500, 900])
    <xarray.Dataset>
    Dimensions:     (ref: 345, wavelength: 201)
    Coordinates:
      * ref         (ref) int64 0 1 2 3 4 5 6 7 ... 337 338 339 340 341 342 343 344
      * wavelength  (wavelength) float64 500.0 502.0 504.0 ... 896.0 898.0 900.0
    Data variables:
        x           (ref) float64 2.057e+05 2.058e+05 ... 2.031e+05 2.03e+05
        y           (ref) float64 8.575e+04 8.58e+04 ... 8.795e+04 8.807e+04
        weight      (ref) float64 11.49 14.13 15.22 14.56 ... 15.21 11.51 8.727
        flux        (ref, wavelength) float64 0.2331 0.231 0.2269 ... 2.213 2.212
    """

    start_wavelength, end_wavelength = wavelength_band

    if start_wavelength > end_wavelength:
        raise ValueError(
            "First input in wavelength_band needs to be smaller than the second."
        )

    data = scene.to_xarray()
    # get dataset with x, y, weight and flux of scene for selected wavelength band.
    selected_data: xr.Dataset = data.sel(
        wavelength=slice(start_wavelength, end_wavelength)
    )

    return selected_data


def integrate_flux(
    flux: xr.DataArray,
) -> xr.DataArray:
    """Integrate flux in photon/(s nm cm2) along the wavelength band in nm and return integrated flux in photon/(s cm2).

    Parameters
    ----------
    flux : xr.DataArray
        Flux. Unit: photon/(s nm cm2).

    Returns
    -------
    integrated_flux : xr.DataArray
        Flux integrated alaong wavelangth band. Unit: photon/(s cm2).

    Examples
    --------
    >>> flux
    <xarray.DataArray 'flux' (ref: 345, wavelength: 201)>
    array([[0.23309256, 0.23099403, 0.22690838, ..., 0.74632666, 0.74254087,
            0.74107508],
           [0.02540123, 0.02539306, 0.02532492, ..., 0.02315146, 0.02275346,
            0.02240911],
           [0.00848251, 0.00849082, 0.00835927, ..., 0.01687145, 0.01677347,
            0.01674323],
           ...,
           [0.0086873 , 0.00874152, 0.0088429 , ..., 0.01047889, 0.01037253,
            0.01032473],
           [0.28250153, 0.27994777, 0.27599212, ..., 0.30550521, 0.30414266,
            0.30421148],
           [3.82353376, 3.86376387, 3.88179622, ..., 2.22050587, 2.21307412,
            2.21216093]])
    Coordinates:
      * ref         (ref) int64 0 1 2 3 4 5 6 7 ... 337 338 339 340 341 342 343 344
      * wavelength  (wavelength) float64 500.0 502.0 504.0 ... 896.0 898.0 900.0
    Attributes:
        units:    ph / (cm2 nm s)
    >>> integrate_flux(flux=flux)
    <xarray.DataArray 'flux' (ref: 345)>
    array([2.13684421e+02, 1.01716326e+01, 5.69110647e+00, 1.17371054e+01,
           2.55767948e+01, 6.91026764e+00, 3.79706245e+00, 1.04048130e+01,
           ...
           6.00547211e+00, 6.19314865e+00, 9.97548328e+00, 5.88036380e+00,
           1.10089431e+01, 1.40244956e+01, 4.23104795e+00, 1.28482189e+02,
           1.18302668e+03])
    Coordinates:
      * ref      (ref) int64 0 1 2 3 4 5 6 7 8 ... 337 338 339 340 341 342 343 344
    Attributes:
        units:    ph / (cm2 s)
    """
    # integrate flux along coordinate wavelength
    integrated_flux = flux.integrate(coord="wavelength")
    integrated_flux.attrs["units"] = str(u.Unit(flux.units) * u.nm)

    return integrated_flux


def convert_flux(
    flux: u.Quantity,
    t_exp: float,
    aperture: float,
) -> u.Quantity:
    """Convert flux in photon/(s cm2) to photon/(s pixel).

    Parameters
    ----------
    flux : u.Quantity
        Flux. Unit: photon/pixel/s/cm2.
    t_exp : float
        Exposure time. Unit: s.
    aperture : float
        Collecting area of the telescope. Unit: m.

    Returns
    -------
    u.Quantity
        Converted flux in photon/s/pixel.

    Examples
    --------
    >>> flux
     <Quantity [2.13684421e+02, 1.01716326e+01, 5.69110647e+00, 1.17371054e+01,
           2.55767948e+01, 6.91026764e+00, 3.79706245e+00, 1.04048130e+01,
           2.67257165e+02, 1.50122471e+01, 5.76777354e+00, 4.18198718e+00,
           ...
           1.18302668e+03] ph / (cm2 s)>
    >>> convert_flux(flux=flux, t_exp=6000 * u.s, aperture=0.1267)
    <Quantity [1.61646841e+08, 7.69458192e+06, 4.30517760e+06, 8.87882234e+06,
           1.93481961e+07, 5.22744209e+06, 2.87238137e+06, 7.87097696e+06,
           ...
           2.10520464e+07, 9.04914055e+06, 4.45057621e+07, 3.61838622e+07,
           4.44387021e+09, 4.28110535e+06, 6.37296766e+06, 9.86548629e+06,
            8.94929654e+08] ph / cm2>
    """

    col_area = np.pi * (aperture * 1e2 / 2) ** 2
    flux_converted = flux * t_exp * col_area

    return flux_converted


def project_objects_to_detector(
    selected_data: xr.Dataset,
    pixel_scale: u.Quantity,
    rows: int,
    cols: int,
) -> np.ndarray:
    """
    Project objects onto detector. Converting scene from arcsec to detector coordinates.

    Parameters
    ----------
    selected_data : xr.Dataset
        Dataset with selected wavelength band and flux to project onto detector.
    pixel_scale : u.Quantity
        Pixel sclae of instrument. Unit: arcsec/pixel.
    rows : int
        Rows of detector.
    cols : int
        Columns of detector.

    Returns
    -------
    projection : np.ndarray
        Projected objects in detector coordinates.

    Examples
    --------
    >>> selected_data
    <xarray.Dataset>
    Dimensions:            (ref: 345, wavelength: 201)
    Coordinates:
      * ref                (ref) int64 0 1 2 3 4 5 6 ... 338 339 340 341 342 343 344
      * wavelength         (wavelength) float64 500.0 502.0 504.0 ... 898.0 900.0
    Data variables:
        x                  (ref) float64 2.057e+05 2.058e+05 ... 2.031e+05 2.03e+05
        y                  (ref) float64 8.575e+04 8.58e+04 ... 8.795e+04 8.807e+04
        weight             (ref) float64 11.49 14.13 15.22 ... 15.21 11.51 8.727
        flux               (ref, wavelength) float64 0.2331 0.231 ... 2.213 2.212
        converted_flux     (ref) float64 1.616e+08 7.695e+06 ... 9.719e+07 8.949e+08
        detector_coords_x  (ref) float64 1.307e+03 1.252e+03 ... 2.748e+03 2.809e+03
        detector_coords_y  (ref) float64 1.454e+03 1.487e+03 ... 2.79e+03 2.859e+03
    >>> project_objects_to_detector(
    ...     selected_data=selected_data,
    ...     pixel_scale=1.65 * u.arcsec / u.pixel,
    ...     rows=4096,
    ...     cols=4132,
    ... )
    array([[0., 0., 0., ..., 0., 0., 0.],
           [0., 0., 0., ..., 0., 0., 0.],
           [0., 0., 0., ..., 0., 0., 0.],
           ...,
           [0., 0., 0., ..., 0., 0., 0.],
           [0., 0., 0., ..., 0., 0., 0.],
           [0., 0., 0., ..., 0., 0., 0.]])
    """
    # we project the stars in the FOV:
    stars_coords = SkyCoord(
        selected_data["x"].values * u.arcsec,
        selected_data["y"].values * u.arcsec,
        frame="icrs",
    )

    # coordinates of telescope pointing
    telescope_ra: u.Quantity = (selected_data["x"].values * u.arcsec).mean()
    telescope_dec: u.Quantity = (selected_data["y"].values * u.arcsec).mean()
    coords_detector = SkyCoord(ra=telescope_ra, dec=telescope_dec, unit="degree")

    # using World Coordinate System (WCS) to convert to pixel
    # more info: https://heasarc.gsfc.nasa.gov/docs/fcg/standard_dict.html
    w = wcs.WCS(naxis=2)

    # define cdelt: coordinate increment along axis
    cdelt = (np.array([-1.0, 1.0]) * pixel_scale).to(u.deg / u.pixel)
    w.wcs.cdelt = cdelt

    # define crpix: coordinate system reference pixel
    crpix = np.array([rows / 2, cols / 2]) * u.pixel
    w.wcs.crpix = crpix

    # define crval: coordinate system value at reference pixel
    w.wcs.crval = [coords_detector.ra.deg, coords_detector.dec.deg]

    # define crota: coordinate system rotation angle
    w.wcs.crota = [0, -0]

    # define ctype: name of the coordinate axis
    w.wcs.ctype = ["RA---TAN", "DEC--TAN"]

    """
    # Other possible method to convert to pixel usingscopesim:
    # https://github.com/AstarVienna/ScopeSim/blob/dev_master/scopesim/optics/image_plane_utils.py#L698
    # gives different result.
    # da = cdelt[0]
    # db = cdelt[1]
    # x0 = crpix[0]
    # y0 = crpix[1]
    # a0 = selected_data["x"].values * u.arcsec
    # b0 = selected_data["y"].values * u.arcsec
    # a = float(telescope_ra) * u.deg
    # b = float(telescope_dec) * u.deg
    # convert stars coordinate to detector coordinates
    # detector_coords_x = x0 + 1. / da * (a - a0)
    # detector_coords_y = y0 + 1. / db * (b - b0)
    """

    detector_coords_x = np.round(
        w.world_to_pixel_values(stars_coords.ra, stars_coords.dec)[0]
    )
    detector_coords_y = np.round(
        w.world_to_pixel_values(stars_coords.ra, stars_coords.dec)[1]
    )
    selected_data["detector_coords_x"] = xr.DataArray(
        detector_coords_x, dims="ref", attrs={"units": "pixel"}
    )
    selected_data["detector_coords_y"] = xr.DataArray(
        detector_coords_y, dims="ref", attrs={"units": "pixel"}
    )

    # make sure that only stars inside the detector
    selected_data_new = selected_data.copy(deep=True)
    selected_data_query = (
        selected_data_new.query(ref="detector_coords_x > 0")
        .query(ref=f"detector_coords_x < {cols}")
        .query(ref=f"detector_coords_y < {rows}")
        .query(ref="detector_coords_y > 0")
    )

    # convert to int
    selected_data2 = selected_data_query.copy(deep=True)
    selected_data2["detector_coords_x"] = selected_data2["detector_coords_x"].astype(
        int
    )
    selected_data2["detector_coords_y"] = selected_data2["detector_coords_y"].astype(
        int
    )

    # get empty array in shape of the detector
    projection = np.zeros([rows, cols])

    # fill in projection of objects in detector coordinates
    for x, group_x in selected_data2.groupby("detector_coords_x"):
        for y, group_y in group_x.groupby("detector_coords_y"):
            projection[int(y), int(x)] = group_y["converted_flux"].values.sum()

    return projection


def simple_aperture(
    detector: Detector,
    pixel_scale: float,
    aperture: float,
    wavelength_band: tuple[float, float],
):
    """Convert scene(photon/s/cm2) to photon.

    Parameters
    ----------
    detector : Detector
        Pyxel detector object.
    pixel_scale : float
        Pixel scale. Unit: arcsec/pixel.
    aperture : float
        Collecting area of the telescope. Unit: m.
    wavelength_band : tuple[float, float]
        Wavelength band. Unit: nm.
    """
    # get dataset for given wavelength and scene object.
    selected_data: xr.Dataset = extract_wavelength(
        scene=detector.scene, wavelength_band=wavelength_band
    )

    # integrate flux
    integrated_flux: xr.DataArray = integrate_flux(flux=selected_data["flux"])

    # get flux in ph/s/cm^2
    flux = u.Quantity(integrated_flux, unit=integrated_flux.units)
    # get time in s
    time = detector.time_step * u.s

    # get flux converted to ph
    converted_flux: u.Quantity = convert_flux(flux=flux, t_exp=time, aperture=aperture)

    # load converted flux to selected dataset
    selected_data["converted_flux"] = xr.DataArray(
        converted_flux, dims="ref", attrs={"units": str(converted_flux.unit)}
    )

    projection = project_objects_to_detector(
        selected_data=selected_data,
        pixel_scale=pixel_scale * u.arcsec / u.pixel,
        rows=detector.geometry.row,
        cols=detector.geometry.col,
    )

    detector.photon.array = projection



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

""":term:`CDM` model calibration with PYGMO.

https://esa.github.io/pagmo2/index.html
"""
import copy
import logging
import math
from collections.abc import Sequence
from copy import deepcopy
from numbers import Number
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Optional, Union

import numpy as np
import pandas as pd
import xarray as xr
from dask.delayed import delayed

from pyxel.calibration import (
    FitRange2D,
    FitRange3D,
    FittingCallable,
    ProblemSingleObjective,
    check_fit_ranges,
    create_processor_data_array,
    read_datacubes,
)
from pyxel.exposure import run_pipeline
from pyxel.inputs import load_dataarray
from pyxel.observation import ParameterValues
from pyxel.pipelines import Processor, ResultId

if TYPE_CHECKING:
    from datatree import DataTree
    from numpy.typing import ArrayLike, NDArray

    from pyxel.exposure import Readout


def build_processors(
    processor: Processor,
    arguments: Sequence[ParameterValues],
) -> Sequence[Processor]:
    max_val, min_val = 0, 1000
    for arg in arguments:
        min_val = min(min_val, len(arg.values))
        max_val = max(max_val, len(arg.values))

    if min_val != max_val:
        logging.warning(
            'The "result_input_arguments" value lists have different lengths! '
            "Some values will be ignored."
        )

    processors: list[Processor] = []
    for i in range(min_val):
        new_processor: Processor = deepcopy(processor)

        step: ParameterValues
        for step in arguments:
            assert step.values != "_"

            value: Union[Literal["_"], str, Number] = step.values[i]

            step.current = value
            new_processor.set(key=step.key, value=step.current)

        processors.append(new_processor)

    return processors


class ModelFittingDataTree(ProblemSingleObjective):
    """Pygmo problem class to fit data with any model in Pyxel."""

    def __init__(
        self,
        processor: Processor,
        variables: Sequence[ParameterValues],
        readout: "Readout",
        simulation_output: ResultId,
        generations: int,
        population_size: int,
        fitness_func: FittingCallable,
        file_path: Path,
        target_fit_range: Union[FitRange2D, FitRange3D],
        out_fit_range: FitRange3D,
        target_filenames: Sequence[Path],
        input_arguments: Optional[Sequence[ParameterValues]] = None,
        weights: Optional[Sequence[float]] = None,
        weights_from_file: Optional[Sequence[Path]] = None,
        pipeline_seed: Optional[int] = None,
    ):
        self._variables: Sequence[ParameterValues] = variables

        self.generations: int = generations
        self.pop: int = population_size
        self.readout: Readout = readout

        self.weighting: Optional[np.ndarray] = None
        self.weighting_from_file: Optional[xr.DataArray] = None
        self.fitness_func: FittingCallable = fitness_func
        self.sim_output: ResultId = simulation_output

        self.file_path: Path = file_path
        self.pipeline_seed: Optional[int] = pipeline_seed

        lower_boundaries, upper_boundaries = self._set_bound()
        self._lower_boundaries: Sequence[float] = lower_boundaries
        self._upper_boundaries: Sequence[float] = upper_boundaries

        if not input_arguments:
            processors: Sequence[Processor] = [deepcopy(processor)]
        else:
            processors = build_processors(
                processor=processor,
                arguments=input_arguments,
            )
        self.param_processor_list: Sequence[Processor] = processors

        num_parameters: int = 0

        var: ParameterValues
        for var in self._variables:
            if isinstance(var.values, list):
                b = len(var.values)
            else:
                b = 1

            num_parameters += b
        self.champion_f_list: np.ndarray = np.zeros((1, 1))
        self.champion_x_list: np.ndarray = np.zeros((1, num_parameters))

        if simulation_output.startswith("data"):
            # Target(s) is/are arrays(s) of unknown number of dimensions.
            # For this reason the file(s) are directly read as 'DataArray' object(s).
            targets_list: Sequence["xr.DataArray"] = [
                load_dataarray(filename) for filename in target_filenames
            ]
            targets: xr.DataArray = xr.concat(targets_list, dim="processor")

            self.targ_fit_range: Union[FitRange2D, FitRange3D, None] = None
            self.sim_fit_range: Optional[FitRange3D] = None
            self.all_target_data: xr.DataArray = targets

            self.target_full_scale: Optional[xr.DataArray] = None

        else:
            self.targ_fit_range = target_fit_range
            self.sim_fit_range = out_fit_range

            if self.readout.time_domain_simulation:
                # TODO: Create a new function 'create_temporal_processor_data_array'
                # Target(s) is/are 3D array(s) of dimensions: 'readout_time', 'y', 'x'
                target_list_3d: Sequence[np.ndarray] = read_datacubes(
                    filenames=target_filenames
                )
                targets_4d: np.ndarray = np.array(target_list_3d)
                num_processors, _, num_y, num_x = targets_4d.shape

                targets = xr.DataArray(
                    target_list_3d,
                    dims=["processor", "readout_time", "y", "x"],
                    coords={
                        "processor": range(num_processors),
                        "y": range(num_y),
                        "x": range(num_x),
                    },
                )

                times = len(targets["readout_time"])
                rows = len(targets["y"])
                cols = len(targets["x"])

                # times, rows, cols = target_list_3d[0].shape
                check_fit_ranges(
                    target_fit_range=target_fit_range,
                    out_fit_range=out_fit_range,
                    rows=rows,
                    cols=cols,
                    readout_times=times,
                )

            else:
                # Get targets from file(s)
                targets = create_processor_data_array(filenames=target_filenames)

                rows = len(targets["y"])
                cols = len(targets["x"])

                check_fit_ranges(
                    target_fit_range=target_fit_range,
                    out_fit_range=out_fit_range,
                    rows=rows,
                    cols=cols,
                )
                self._configure_weights(
                    weights=weights,
                    weights_from_file=weights_from_file,
                )

            self.all_target_data = targets.isel(indexers=target_fit_range.to_dict())
            self.target_full_scale = targets

    def get_bounds(self) -> tuple[Sequence[float], Sequence[float]]:
        """Get the box bounds of the problem (lower_boundary, upper_boundary).

        It also implicitly defines the dimension of the problem.

        Returns
        -------
        tuple of lower boundaries and upper boundaries
        """
        return self._lower_boundaries, self._upper_boundaries

    def _configure_weights(
        self,
        weights: Optional[Sequence[float]] = None,
        weights_from_file: Optional[Sequence[Path]] = None,
    ) -> None:
        """TBW.

        Parameters
        ----------
        weights
        weights_from_file
        """
        if weights_from_file is not None:
            assert self.targ_fit_range is not None

            if self.readout.time_domain_simulation:
                weights_list: Sequence[np.ndarray] = read_datacubes(weights_from_file)
                weights_data_array = xr.DataArray(
                    weights_list, dims=["processor", "readout_time", "y", "x"]
                )
            else:
                weights_data_array = create_processor_data_array(
                    filenames=weights_from_file
                )

            self.weighting_from_file = weights_data_array.isel(
                indexers=self.targ_fit_range.to_dict()
            )

        elif weights is not None:
            self.weighting = np.array(weights)

    def _set_bound(self) -> tuple[Sequence[float], Sequence[float]]:
        lbd: list[float] = []
        ubd: list[float] = []

        var: ParameterValues
        for var in self._variables:
            assert var.boundaries is not None  # TODO: Fix this

            if var.values == "_":
                assert var.boundaries.shape == (2,)  # TODO: Fix this

                low_val: float
                high_val: float
                low_val, high_val = var.boundaries

                if var.logarithmic:
                    low_val = math.log10(low_val)
                    high_val = math.log10(high_val)

                lbd += [low_val]
                ubd += [high_val]

            elif isinstance(var.values, Sequence) and all(
                x == "_" for x in var.values[:]
            ):
                if var.boundaries.ndim == 1:
                    low_val, high_val = var.boundaries

                    low_values: NDArray[np.float_] = np.array(
                        [low_val] * len(var.values)
                    )
                    high_values: NDArray[np.float_] = np.array(
                        [high_val] * len(var.values)
                    )

                elif var.boundaries.ndim == 2:
                    low_values = var.boundaries[:, 0]
                    high_values = var.boundaries[:, 1]
                else:
                    raise NotImplementedError

                if var.logarithmic:
                    low_values = np.log10(low_values)
                    high_values = np.log10(high_values)

                lbd += low_values.tolist()
                ubd += high_values.tolist()

            else:
                raise ValueError(
                    'Character "_" (or a list of it) should be used to '
                    "indicate variables need to be calibrated"
                )

        return lbd, ubd

    def _get_simulated_data(self, data: "DataTree") -> "xr.DataArray":
        """Extract 2D data from a processor."""
        import xarray as xr

        if self.sim_output not in (
            "image",
            "signal",
            "pixel",
        ) and not self.sim_output.startswith("data"):
            raise NotImplementedError(
                f"Simulation mode: {self.sim_output!r} not implemented"
            )

        simulated_data = data[self.sim_output]
        if not isinstance(simulated_data, xr.DataArray):
            raise TypeError("Expected a 'DataArray'")

        if self.sim_fit_range is not None:
            simulated_data = simulated_data.isel(indexers=self.sim_fit_range.to_dict())

        return simulated_data

    def _calculate_fitness(
        self,
        simulated_data: "xr.DataArray",
        target_data: "xr.DataArray",
        weighting: Optional[np.ndarray] = None,
    ) -> float:
        if self.sim_output.startswith("data"):
            assert simulated_data.ndim == target_data.ndim

            # Create 'simulated_renamed' with the same dimensions as 'target_data'
            renamed_dimensions = dict(zip(simulated_data.dims, target_data.dims))
            simulated_renamed: xr.DataArray = simulated_data.rename(renamed_dimensions)

            # 'simulated_interpolated' has the same coordinates as 'target_data'
            simulated_interpolated = simulated_renamed.interp_like(target_data)

            simulated_2d = np.array(simulated_interpolated, dtype=float)

        else:
            simulated_2d = np.array(simulated_data, dtype=float)

        target_2d = np.array(target_data, dtype=float)

        if weighting is not None:
            factor = weighting
        else:
            factor = np.ones_like(target_2d)

        weighting_2d = np.array(factor, dtype=float)

        fitness: float = self.fitness_func(
            simulated=simulated_2d,
            target=target_2d,
            weighting=weighting_2d,
        )

        return fitness

    # TODO: If possible, use 'numba' for this method
    def fitness(self, decision_vector_1d: np.ndarray) -> Sequence[float]:
        """Call the fitness function, elements of parameter array could be logarithmic values.

        Parameters
        ----------
        decision_vector_1d : array_like
            A 1d decision vector.

        Returns
        -------
        sequence
            The fitness of the input decision vector (concatenating the objectives,
            the equality and the inequality constraints)
        """
        # TODO: Fix this
        if self.pop is None:
            raise NotImplementedError("'pop' is not initialized.")

        try:
            # TODO: Use directory 'logging.'
            logger = logging.getLogger("pyxel")
            prev_log_level = logger.getEffectiveLevel()

            parameter_1d = self.convert_to_parameters(decision_vector_1d)
            # TODO: deepcopy is not needed. Check this
            processor_list: Sequence[Processor] = self.param_processor_list

            overall_fitness: float = 0.0
            for processor_id, (processor, target_data) in enumerate(
                zip(processor_list, self.all_target_data)
            ):
                processor = self.update_processor(
                    parameter=parameter_1d, processor=processor
                )

                logger.setLevel(logging.WARNING)

                data_tree: "DataTree" = run_pipeline(
                    processor=processor,
                    readout=self.readout,
                    pipeline_seed=self.pipeline_seed,
                )

                logger.setLevel(prev_log_level)

                simulated_data: "xr.DataArray" = self._get_simulated_data(
                    data=data_tree
                )

                weighting: Optional[np.ndarray] = None

                if self.weighting is not None:
                    weighting = np.full(
                        shape=(
                            processor.detector.geometry.row,
                            processor.detector.geometry.col,
                        ),
                        fill_value=self.weighting[processor_id],
                    )
                elif self.weighting_from_file is not None:
                    weighting_data: xr.DataArray = self.weighting_from_file.isel(
                        processor=processor_id
                    )
                    weighting = weighting_data.to_numpy()

                overall_fitness += self._calculate_fitness(
                    simulated_data=simulated_data,
                    target_data=target_data,
                    weighting=weighting,  # TODO: 'weighting' should be a 'DataArray'
                )

        except Exception:
            logging.exception(
                "Catch an exception in 'fitness' for ModelFitting: %r.", self
            )
            raise

        return [overall_fitness]

    def convert_to_parameters(self, decisions_vector: "ArrayLike") -> np.ndarray:
        """Convert a decision version from Pygmo2 to parameters.

        Parameters
        ----------
        decisions_vector : array_like
            It could a 1D or 2D array.

        Returns
        -------
        array_like
            Parameters
        """
        parameters = np.array(decisions_vector)

        a = 0
        for var in self._variables:
            b = 1
            if isinstance(var.values, list):
                b = len(var.values)
            if var.logarithmic:
                start = a
                stop = a + b
                parameters[..., start:stop] = np.power(10, parameters[..., start:stop])
            a += b

        return parameters

    def apply_parameters(
        self, processor: Processor, parameter: np.ndarray
    ) -> "DataTree":
        """Create a new ``Processor`` with new parameters."""
        new_processor = self.update_processor(parameter=parameter, processor=processor)

        data_tree: "DataTree" = run_pipeline(
            processor=new_processor,
            readout=self.readout,
            pipeline_seed=self.pipeline_seed,
        )

        return data_tree

    def apply_parameters_to_processors(
        self, parameters: "xr.DataArray"
    ) -> pd.DataFrame:
        """TBW."""
        assert "island" in parameters.dims
        assert "param_id" in parameters.dims

        lst = []
        for id_processor, processor in enumerate(self.param_processor_list):
            delayed_processor = delayed(processor)

            for idx_island, params_array in parameters.groupby("island"):
                params: np.ndarray = params_array.data  # type

                result_datatree: DataTree = delayed(self.apply_parameters)(
                    processor=delayed_processor, parameter=params
                )

                lst.append(
                    {
                        "island": idx_island,
                        "id_processor": id_processor,
                        "data_tree": result_datatree,
                    }
                )

        df = pd.DataFrame(lst).sort_values(["island", "id_processor"])

        return df

    def update_processor(
        self, parameter: np.ndarray, processor: Processor
    ) -> Processor:
        """TBW."""
        new_processor = copy.deepcopy(processor)
        a, b = 0, 0
        for var in self._variables:
            if var.values == "_":
                b = 1
                new_processor.set(key=var.key, value=parameter[a])
            elif isinstance(var.values, list):
                b = len(var.values)

                start = a
                stop = a + b
                new_processor.set(key=var.key, value=parameter[start:stop])
            a += b
        return new_processor



from pathlib import Path
from typing import Union

import numpy as np
import pytest

from pyxel.detectors import CCD, CCDGeometry, Characteristics, Environment
from pyxel.models.charge_collection import fixed_pattern_noise
from pyxel.models.charge_collection.fixed_pattern_noise import compute_simple_prnu


@pytest.fixture
def ccd_5x5() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(quantum_efficiency=0.9),
    )
    return detector


@pytest.fixture
def valid_noise_path(
    tmp_path: Path,
) -> str:
    """Create valid 2D file on a temporary folder."""
    data_2d = (
        np.array(
            [
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0],
            ]
        )
        * 0.5
    )

    final_path = f"{tmp_path}/noise.npy"
    np.save(final_path, arr=data_2d)

    return final_path


def test_fixed_pattern_noise_valid_path(
    ccd_5x5: CCD, valid_noise_path: Union[str, Path]
):
    """Test function fixed_pattern_noise with valid path inputs."""

    detector = ccd_5x5

    array = np.ones((5, 5))
    detector.pixel.array = array
    target = array * 0.5

    fixed_pattern_noise(detector=detector, filename=valid_noise_path)

    np.testing.assert_array_almost_equal(detector.pixel.array, target)


def test_compute_simple_prnu(ccd_5x5: CCD):
    """Test function 'compute_simple_prnu'."""
    detector = ccd_5x5
    shape = detector.geometry.shape
    quantum_efficiency = detector.characteristics.quantum_efficiency
    fixed_pattern_noise_factor = 0.01

    prnu_2d = compute_simple_prnu(shape, quantum_efficiency, fixed_pattern_noise_factor)
    np.testing.assert_equal(prnu_2d.shape, shape)
    assert np.all(prnu_2d >= quantum_efficiency)


def test_fixed_pattern_noise_valid(ccd_5x5: CCD):
    """Test model fixed_pattern_noise with valid fpn inputs."""
    detector = ccd_5x5
    fixed_pattern_noise(detector=detector, fixed_pattern_noise_factor=0.01)


def test_fpn_raises(ccd_5x5: CCD, valid_noise_path: Union[str, Path]):
    """Test model fixed_pattern_noise when generating an error."""
    detector = ccd_5x5
    invalid_noise_path = "noise.npy"
    with pytest.raises(ValueError, match="filename or fixed_pattern_noise_factor"):
        fixed_pattern_noise(
            detector=detector,
            filename=valid_noise_path,
            fixed_pattern_noise_factor=0.01,
        )
    with pytest.raises(ValueError, match="filename or fixed_pattern_noise_factor"):
        fixed_pattern_noise(detector=detector)

    with pytest.raises(FileNotFoundError, match="Cannot find folder"):
        fixed_pattern_noise(detector=detector, filename=invalid_noise_path)



# Copyright or © or Copr. Thibault Pichon, CEA Paris-Saclay (2023)
#
# thibault.pichon@cea.fr
#
# This file is part of the Pyxel general simulator framework.
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


"""Pyxel photon generator models for ARIEL AIRS."""

import logging
from pathlib import Path

import numpy as np
import pandas as pd
from astropy import units as u
from astropy.io import ascii, fits
from astropy.units import Quantity
from scipy.integrate import cumtrapz

from pyxel.detectors import Detector


def read_star_flux_from_file(
    filename: str,
) -> tuple[np.ndarray, np.ndarray]:
    """Read star flux file.

    # TODO: Read the unit from the text file.

    Parameters
    ----------
    filename: type string,
        Name of the target file. Different extension can be considered.

    Returns
    -------
    wavelength: type array 1D
        Wavelength associated with the flux of the star.
    flux: type array 1D
        Flux of the target considered, in  ph/s/m2/µm.
    """
    extension = Path(filename).suffix
    if extension == ".txt":
        wavelength, flux = np.loadtxt(filename).T
        # set appropriate unit here um
        wavelength = wavelength * u.micron
        # set appropriate units
        flux = flux * u.photon / u.s / u.m / u.m / u.micron

    elif extension == ".ecsv":
        data = ascii.read(filename)
        flux = data["flux"].data * data["flux"].unit
        wavelength = data["wavelength"].data * data["wavelength"].unit

    elif extension == ".dat":
        """ExoNoodle file"""
        df_topex = pd.read_csv(filename, header=11)
        wavelength, flux = np.zeros(len(df_topex["wavelength flux "])), np.zeros(
            len(df_topex["wavelength flux "])
        )
        for i in range(len(df_topex["wavelength flux "])):
            wavelength[i] = float(df_topex["wavelength flux "][i].split("        ")[0])
            conv = 1.51 * 1e3 * 1e4 / wavelength[i]
            flux[i] = (
                float(df_topex["wavelength flux "][i].split("        ")[1])
                * conv
                * 1e-6
            )  # Attention aux unités
        wavelength = wavelength * u.micron
        flux = flux * u.photon / u.s / u.m / u.m / u.micron
    else:
        logging.debug("ERROR while converting, extension not readable")

    return wavelength, flux


def convert_flux(
    wavelength: np.ndarray,
    flux: Quantity,
    telescope_diameter_m1: float,
    telescope_diameter_m2: float,
) -> np.ndarray:
    """Convert the flux of the target in ph/s/µm.

    Parameters
    ----------
    wavelength: 1D array
        Wavelength sampling of the considered target.
    flux: 1D array
        Flux of the target considered in ph/s/m2/µm.
    telescope_diameter_m1: float
        Diameter of the M1 mirror of the TA in m.
    telescope_diameter_m2: float
        Diameter of the M2 mirror of the TA in m.

    Returns
    -------
    conv_flux: 1D array
        Flux of the target considered in ph/s/µm.
    """

    logging.debug("Incident photon flux is being converted into ph/s/um.")
    # use of astropy code
    flux.to(
        u.photon / u.m**2 / u.micron / u.s,
        equivalencies=u.spectral_density(wavelength),
    )

    diameter_m2 = (
        telescope_diameter_m2 * u.meter
    )  # TODO: define a class to describe the optic of the telescope ?
    diameter_m1 = telescope_diameter_m1 * u.meter
    collecting_area = np.pi * diameter_m1 * diameter_m2 / 4
    conv_flux = np.copy(flux) * collecting_area

    return conv_flux


def compute_bandwidth(
    psf_wavelength: Quantity,
) -> tuple[Quantity, Quantity]:
    """Compute the bandwidth for non-even distributed values.

    First we put the poles, each pole is at the center of the previous wave and the next wave.
    We add the first pole and the last pole using symmetry. We get nw+1 poles

    Parameters
    ----------
    psf_wavelength : Quantity
       PSF object.

    Returns
    -------
    bandwidth : quantity array
        Bandwidth. Usually in microns.
    all_poles : quantity array
        Pole wavelengths. Dimension: nw
    """
    poles = (psf_wavelength[1:] + psf_wavelength[:-1]) / 2
    first_pole = psf_wavelength[0] - (psf_wavelength[1] - psf_wavelength[0]) / 2
    last_pole = psf_wavelength[-1] + (psf_wavelength[-1] - psf_wavelength[-2]) / 2
    all_poles = np.concatenate(([first_pole], poles, [last_pole]))
    bandwidth = all_poles[1:] - all_poles[:-1]

    return bandwidth, all_poles


def integrate_flux(
    wavelength: Quantity,
    flux: Quantity,
    psf_wavelength: Quantity,
) -> np.ndarray:
    """Integrate flux on each bin around the psf.

    The trick is to integrate first, and interpolate after (and not vice-versa).

    Parameters
    ----------
    wavelength : quantity array
        Wavelength. Unit: usually micron. Dimension: small_n.
    flux : quantity array
        Flux. Unit: ph/s/m2/micron. Dimension: big_n.
    psf_wavelength : quantity array
        Point Spead Function per wavelength.

    Returns
    -------
    flux : quantity array
        Flux. UNit: photon/s. Dimension: nw.
    """

    logging.debug("Integrate flux on each bin around the psf...")

    # Set the parameters of the function
    bandwidth, all_poles = compute_bandwidth(psf_wavelength)

    # Cumulative count
    cum_sum = cumtrapz(flux.value, wavelength.value, initial=0.0)

    # self.wavelength has to quantity: value and units
    # interpolate over psf wavelength
    cum_sum_interp = np.interp(all_poles, wavelength, cum_sum)

    # Compute flux over psf spectral bin
    flux_int = cum_sum_interp[1:] - cum_sum_interp[:-1]

    # Update flux matrix
    flux_int = flux_int * flux.unit * wavelength.unit
    flux = np.copy(flux_int)

    return flux


# def multiply_by_transmission(psf, transmission_dict: Mapping[str, Any]) -> None:
#     """The goal of this function is to take into account the flux of the incident star"""
#     for t in transmission_dict.keys():
#         if "M" in t:
#             f = interpolate.interp1d(
#                 transmission_dict[t]["wavelength"],
#                 transmission_dict[t]["reflectivity_eol"],
#             )
#         else:
#             f = interpolate.interp1d(
#                 transmission_dict[t]["wavelength"],
#                 transmission_dict[t]["transmission_eol"],
#             )
#         flux = np.copy(flux) * f(psf.psf_wavelength)
#


def read_psf_from_fits_file(
    filename: str,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Read psf files depending on simulation and instrument parameters.

    Parameters
    ----------
    filename : str
        Name of the .fits file.

    Returns
    -------
    psf_datacube : ndarray
        3D array, PSF for each wavelength saved as array, one for each wavelength
        waves (1D array) wavelength. Unit: um.
    psf_wavelength : ndarray
        1D array, wavelengths. Unit: um.
    line_pos_psf : ndarray
        1D array, x position of the PSF at each wavelength.
    col_pos_psf : ndarray
        1D array, y position of the PSF at each wavelength.
    """
    # Open fits
    with fits.open(filename) as hdu:
        psf_datacube, table = hdu[0].data, hdu[1].data

        # Position of the PSF on AIRS window along line
        line_psf_pos = (table["x_centers"]).astype(int)

        # Position of the PSF on AIRS window along col
        col_psf_pos = (table["y_centers"]).astype(int)

        # Wavelength
        psf_wavelength = table["waves"] * u.micron

    logging.debug(
        "PSF Datacube %r %r %r",
        psf_datacube.shape,
        psf_datacube.min(),
        psf_datacube.max(),
    )
    logging.debug(
        "PSF Wavelength %r %r %r",
        psf_wavelength.shape,
        psf_wavelength.min(),
        psf_wavelength.max(),
    )

    return psf_datacube, psf_wavelength, line_psf_pos, col_psf_pos


def project_psfs(
    psf_datacube_3d: np.ndarray,
    line_psf_pos_1d: Quantity,
    col_psf_pos,
    flux: Quantity,
    row: int,
    col: int,
    expand_factor: int,
) -> tuple[Quantity, Quantity]:
    """Project each psf on a (n_line_final * self.zoom, n_col_final * self.zoom) pixel image.

    n_line_final, n_col_final = corresponds to window size. It varies with the channel.

    Parameters
    ----------
    psf_datacube_3d : numpy array
        Dimension (nw, n_line, n_col).
    line_psf_pos_1d : numpy array
        Line position of the center of the PSF along AIRS window.
    col_psf_pos : numpy array
        Column position of the center of PSF along AIRS window.
    flux : Quantity
        Unit electron/s dimension big_n.
    row :

    col :

    expand_factor :

    Returns
    -------
    result : Quantity, Quantity
        Dimension ny, nx, spectral image in e-/s. Shape = detector shape.

    """
    nw, n_line, n_col = psf_datacube_3d.shape  # Extract shape of the PSF
    half_size_col, half_size_line = n_col // 2, n_line // 2  # Half size of the PSF

    # photon_incident = np.zeros((detector.geometry.row * expend_factor, detector.geometry.col * expend_factor))
    # photoelectron_generated = np.zeros((detector.geometry.row * expend_factor, detector.geometry.col * expend_factor))
    photon_incident = np.zeros((row * expand_factor, col * expand_factor))
    photoelectron_generated = np.zeros((row * expand_factor, col * expand_factor))

    col_win_middle, line_win_middle = int(col_psf_pos.mean()), int(
        line_psf_pos_1d.mean()
    )

    for i in np.arange(nw):  # Loop over the wavelength
        # Resize along line dimension
        line1 = (
            line_psf_pos_1d[i]
            - line_win_middle
            + row * expand_factor // 2
            - half_size_line
        )
        line2 = (
            line_psf_pos_1d[i]
            - line_win_middle
            + row * expand_factor // 2
            + half_size_line
        )
        # Resize along col dimension
        col1 = (
            col_psf_pos[i] - col_win_middle + col * expand_factor // 2 - half_size_col
        )
        col2 = (
            col_psf_pos[i] - col_win_middle + col * expand_factor // 2 + half_size_col
        )
        # Derive the amount of photon incident on the detector
        photon_incident[line1:line2, col1:col2] = (
            photon_incident[line1:line2, col1:col2]
            + psf_datacube_3d[i, :, :] * flux[i].value
        )

        # TODO: Here take into account the QE map of the detector, and its dependence with wavelength
        # QE of the detector has to be sampled with the same resolution as the PSF is sampled
        qe = 0.65
        # qe_detector[i]
        photoelectron_generated[line1:line2, col1:col2] = (
            photon_incident[line1:line2, col1:col2].copy() * qe
        )

    # This is the amount of photons incident on the detector
    photon_incident = photon_incident * flux.unit

    # This is the amount of photons incident on the detector
    photoelectron_generated = photoelectron_generated * u.electron

    return (rebin_2d(photon_incident, expand_factor)), (
        rebin_2d(photoelectron_generated, expand_factor)
    )


def rebin_2d(
    data: np.ndarray,
    expand_factor: int,
) -> np.ndarray:
    """Rebin as idl.

    Each pixel of the returned image is the sum of zy by zx pixels of the input image.

    Based on:       Rene Gastaud, 13 January 2016
    https://codedump.io/share/P3pB13TPwDI3/1/resize-with-averaging-or-rebin-a-numpy-2d-array
    2017-11-17 : RG and Alan O'Brien  compatibility with python 3 bug452

    Parameters
    ----------
    data : ndarray
        Data with 2 dimensions (image): ny, nx.
    expand_factor : tuple
        Expansion factor is a tuple of 2 integers: zy, zx.

    Returns
    -------
    result : ndarray
        Shrunk in dimension ny/zy, nx/zx.


    Example
    -------
    a = np.arange(48).reshape((6,8))
               rebin2d( a, [2,2])

    """

    # In case asymmetrical zoom is used
    zoom = [expand_factor, expand_factor]

    final_shape = (
        int(data.shape[0] // zoom[0]),
        zoom[0],
        int(data.shape[1] // zoom[1]),
        zoom[1],
    )
    logging.debug("final_shape ", final_shape)
    result = data.reshape(final_shape).sum(3).sum(1)

    return result


def wavelength_dependence_airs(
    detector: Detector,
    psf_filename: str,
    target_filename: str,
    telescope_diameter_m1: float,
    telescope_diameter_m2: float,
    expand_factor: int,
    time_scale: float = 1.0,
) -> None:
    """Generate the photon over the array according to a specific dispersion pattern (ARIEL-AIRS).

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    psf_filename : string
        The location and the filename where the PSFs are located.
    target_filename : string
        The location and the filename of the target file used in the simulation.
    telescope_diameter_m1 : float
        Diameter of the M1 mirror of the TA in m.
    telescope_diameter_m2 : float
        Diameter of the M2 mirror of the TA in m.
    expand_factor : int
        Expansion factor used.
    time_scale : float
        Time scale in seconds.
    """
    if not isinstance(expand_factor, int):
        raise TypeError("Expecting a 'int' type for 'expand_factor'.")

    if expand_factor <= 0:
        raise ValueError("Expecting a positive value for 'expand_factor'.")

    # Extract information from the PSF
    psf_datacube, psf_wavelength, line_psf_pos, col_psf_pos = read_psf_from_fits_file(
        filename=psf_filename
    )
    # plt.figure()
    # plt.pcolormesh(psf_datacube[10, :, :])
    # plt.title(psf_wavelength[10])

    # Read flux from the fits file
    target_wavelength, target_flux = read_star_flux_from_file(filename=target_filename)

    # convert the flux by multiplying by the area of the detector
    # telescope_diameter_m1, telescope_diameter_m2 = (
    #     1.1,
    #     0.7,
    # )  # in m, TODO to be replaced by Telescope Class ?

    target_conv_flux = convert_flux(
        wavelength=target_wavelength,
        flux=target_flux,
        telescope_diameter_m1=telescope_diameter_m1,
        telescope_diameter_m2=telescope_diameter_m2,
    )
    # Integrate the flux over the PSF spectral bin
    integrated_flux = integrate_flux(
        target_wavelength, target_conv_flux, psf_wavelength
    )  # The flux is now sample similarly to the PSF

    # The Flux can be multiplied here by the optical elements
    # multiply_by_transmission(psf, transmission_dict)
    # #TODO add class to take into account the transmission of the instrument
    # Project the PSF onto the Focal Plane, to get the detector image

    # row, col = 130, 64  # Could be replaced
    # Expend factor used: expand_factor = 18
    photon_incident, photo_electron_generated = project_psfs(
        psf_datacube_3d=psf_datacube,
        line_psf_pos_1d=line_psf_pos,
        col_psf_pos=col_psf_pos,
        flux=integrated_flux,
        row=detector.geometry.row,
        col=detector.geometry.col,
        expand_factor=expand_factor,
    )
    # Add the result to the photon array structure
    time_step = 1.0  # ?

    photon_array = photo_electron_generated * (time_step / time_scale)
    assert photon_array.unit == "electron"

    detector.photon.array += np.array(photon_array)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Simple model to compute basic statistics."""
from collections.abc import Sequence
from typing import TYPE_CHECKING, Literal, Union

import xarray as xr
from datatree import DataTree

from pyxel.detectors import Detector

if TYPE_CHECKING:
    from pyxel.data_structure import Image, Photon, Pixel, Signal


def compute_statistics(
    data_array: xr.DataArray,
    absolute_time: xr.DataArray,
    dimensions: Union[str, Sequence[str]] = ("x", "y"),
) -> xr.Dataset:
    """Compute basic statistics and save it into a Dataset.

    Parameters
    ----------
    data_array : DataArray
        Input array which contains the data.
    absolute_time : DataArray
        Absolute time(s) of the data.
    dimensions : str or Sequence of str
        Dimensions, where the statistics should be computed on. Default is ("x", "y").

    Returns
    -------
    Dataset
    """

    var = data_array.var(dim=dimensions)
    mean = data_array.mean(dim=dimensions)
    min_array = data_array.min(dim=dimensions)
    max_array = data_array.max(dim=dimensions)
    count = data_array.count(dim=dimensions)

    dataset = xr.Dataset()
    dataset["var"] = var
    dataset["mean"] = mean
    dataset["min"] = min_array
    dataset["max"] = max_array
    dataset["count"] = count

    dataset = dataset.expand_dims(dim="time").assign_coords(time=absolute_time)

    return dataset


def statistics(
    detector: Detector,
    data_structure: Literal["pixel", "photon", "signal", "image", "all"] = "all",
) -> None:
    """Compute basic statistics.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    data_structure : Literal
        Keyword to choose data structure. Can be any from:
        ("pixel", "photon", "signal", "image", "all").
        Default is "all" and computes the statistics on "pixel", "photon", "signal" and "image".

    """
    if data_structure == "all":
        names = ["pixel", "photon", "signal", "image"]
    else:
        names = [data_structure]

    parent: str = "/statistics"
    parent_partial: str = f"{parent}/partial"

    for name in names:
        # Extract data from 'detector'
        data_bucket: Union[Pixel, Photon, Signal, Image] = getattr(detector, name)
        data_array: xr.DataArray = data_bucket.to_xarray(dtype=float)

        # Get current absolute time
        absolute_time = xr.DataArray(
            [detector.absolute_time],
            dims=["time"],
            attrs={"units": "s"},
        )

        # Get statistics data
        statistics: xr.Dataset = compute_statistics(
            data_array=data_array, absolute_time=absolute_time
        )

        key: str = f"{parent}/{name}"
        key_partial: str = f"{parent_partial}/{name}"

        try:
            _ = detector.data[key_partial]
        except KeyError:
            has_key_partial = False
        else:
            has_key_partial = True

        if not has_key_partial:
            data_tree: DataTree = DataTree(statistics)
        else:
            # Concatenate data
            previous_datatree = detector.data[key_partial]
            data_tree = previous_datatree.combine_first(statistics)  # type: ignore

        if detector.pipeline_count == (detector.num_steps - 1):
            detector.data[key] = data_tree
        else:
            detector.data[key_partial] = data_tree

    # This is the last step and there is at least two steps
    if detector.num_steps > 1 and (detector.pipeline_count == (detector.num_steps - 1)):
        detector.data[parent_partial].orphan()
